"""
    Swagger Words
    ~~~~~~~~~~~~~

    Python friendly aliases to reserved Swagger words.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from __future__ import unicode_literals

additional_properties = "additionalProperties"
all_of = "allOf"
allow_empty_value = "allowEmptyValue"
any_of = "anyOf"
api_key = "apiKey"
array = "array"
basic = "basic"
binary = "binary"
body = "body"
boolean = "boolean"
byte = "byte"
collection_format = "collectionFormat"
components = "components"
consumes = "consumes"
content = "content"
csv = "csv"
date = "date"
date_time = "date-time"
default = "default"
definitions = "definitions"
description = "description"
double = "double"
enum = "enum"
external_docs = "externalDocs"
exclusive_maximum = "exclusiveMaximum"
exclusive_minimum = "exclusiveMinimum"
explode = "explode"
float_ = "float"
form = "form"
format_ = "format"
header = "header"
host = "host"
in_ = "in"
info = "info"
integer = "integer"
int32 = "int32"
int64 = "int64"
items = "items"
max_items = "maxItems"
max_length = "maxLength"
max_properties = "maxProperties"
maximum = "maximum"
min_items = "minItems"
min_length = "minLength"
min_properties = "minProperties"
minimum = "minimum"
multi = "multi"
multiple_of = "multipleOf"
name = "name"
null = "null"
nullable = "x-nullable"
number = "number"
oauth2 = "oauth2"
object_ = "object"
one_of = "oneOf"
openapi = "openapi"
operation_id = "operationId"
parameters = "parameters"
password = "password"
path = "path"
paths = "paths"
pattern = "pattern"
produces = "produces"
properties = "properties"
query = "query"
ref = "$ref"
request_body = "requestBody"
required = "required"
responses = "responses"
schema = "schema"
schemas = "schemas"
schemes = "schemes"
security = "security"
security_definitions = "securityDefinitions"
security_schemes = "securitySchemes"
servers = "servers"
simple = "simple"
string = "string"
style = "style"
swagger = "swagger"
tags = "tags"
title = "title"
type_ = "type"
unique_items = "uniqueItems"
url = "url"
uuid = "uuid"
variables = "variables"
version = "version"
